from __future__ import print_function
from . import utils
import os
import math
from shapely.geometry import Point, LineString, Polygon, MultiLineString
from scipy import spatial
import geopandas as gpd
import numpy as np
from datetime import datetime

HOMEDIR = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))


class SloTransformation:

    methods = ["triangle", "24regions"]

    not_implemented_methods = ["1region", "3regions", "7regions"]

    def __init__(self,from_crs, method="triangle",params=None):

        if method not in self.methods:
            if method in self.not_implemented_methods:
                raise NotImplementedError("Method {} is not yet implemented! So far it can only be one of {}!".format(self.method,self.methods))
            else:
                raise IOError("Type can only be one of {}!".format(",".join(self.methods)))

        if from_crs not in ["d96", "d48"]:
            raise IOError("From_crs argument can only be d96 or d48!")

        self.method = method
        self.params = params
        self.from_crs = from_crs

        self.shpfile = os.path.join(HOMEDIR, "pysitra","static", "{}_parameters_from_{}.shp".format(self.method, self.from_crs))


        if self.params != None:  # Custom parameters! Error check parameteres
            if self.method in ["1region","3regions","7regions"]:
                if len(params) != 7:
                    raise IOError("You specified manual parameters: {}. Method {} requires 7 parameters in order 'cx;cy;cz;s;rx;ry;rz' (3x translation, 1x scale, 3x rotation)!"
                                  .format(self.params,self.method,len))

            if self.method == "24regions":
                if len(params) != 4:
                    raise IOError("You specified manual parameters: {}. Method 24regions requires 4 parameters in order "
                                  "'scale;rotation;transx;transy'!".format(self.params))

            if self.method == "triangle":
                if len(params) != 6:
                    raise IOError(
                        "You specified manual parameters: {}. Method triangle requires 6 parameters in order 'a;b;c;d;e;f' "
                        "(scalex,scaley,rotationx,rotationy,transx,transy)!".format(self.params))

        #Prepare spatial data for faster execution of transform function! (Reading shapefiles and preparing special objects
        if self.method != "1region":
            # 1region doesnt have shapefile, parameters are hardcoded!
            self.df_parameters = gpd.read_file(self.shpfile)

        if self.method == "triangle":
            self.reference_points = np.array([(p.x, p.y) for p in self.df_parameters["geometry"].values.tolist()])

            # create a KDTre object from a reference points
            self.KDTree_object = spatial.KDTree(self.reference_points)


    def transform(self,points):

        # INPUT ERROR CHECK:
        if not isinstance(points, list):
            points = [points]

        if isinstance(points[0], Point):
            points = [(p.x, p.y) for p in points]

        new_points = []

        # AFFINE TRIANGLE 2R 6PARAMETRIC
        if self.method == "triangle":

            for point in points:
                index_of_closest = self.KDTree_object.query(point)[1]
                if self.params:
                    params = self.params
                else:
                    params = self.df_parameters.reset_index().ix[
                        index_of_closest, "a b c d e f".split(" ")].values.tolist()
                # https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array

                x_new, y_new = trans_2R_6params(point[0], point[1], params)
                new_points.append(Point(round(x_new, 2), round(y_new, 2)))

        # HELMERT 3R 7 PARAMETRIC
        elif self.method in ["1region", "3regions", "7regions"]:

            # doloci parametre
            if self.method == "1region":
                if self.params:
                    params = self.params
                else:
                    if self.from_crs == "d48":
                        # from d48 to d96
                        params = [409.545, 72.164, 486.872, 17.919665, -3.085957, -5.469110, 11.020289]
                    else:
                        # from d96 to d48
                        params = [-409.520, -72.192, -486.872, -17.919456, 3.086250, 5.468945, -11.020370]
            else:
                closest_polygons = utils.closest_element_to_given_points(points, self.df_parameters)

            for i, point in enumerate(points):
                if self.method == "1region":
                    x_new, y_new, _ = trans_3R_7params(point[0], point[1], 0, params=params)
                else:
                    params = self.df_parameters.ix[closest_polygons[i], "cx cy cz s rx ry rz".split(" ")]  # daj ven "id field"
                    #TODO:Impletement 3D points!
                    x_new, y_new, _ = trans_3R_7params(point[0], point[1], 0, params=params)

                new_points.append(Point(round(x_new,2), round(y_new,2)))

        #AFFINE 2R 4parametric
        elif self.method == "24regions":
            closest_polygons = utils.closest_element_to_given_points(points, self.df_parameters)

            for i, point in enumerate(points):
                if self.params:
                    params = self.params
                else:
                    params = self.df_parameters.ix[closest_polygons[i], ["merilo","zasuk","easting","northing"]]
                x_new, y_new = trans_2R_4params(point[0], point[1], params=params)

                new_points.append(Point(round(x_new, 2), round(y_new, 2)))

        if len(new_points) == 1:
            return new_points[0]

        return new_points

def trans_2R_6params(x,y,params=()):
    #parameters in order a,b,c,d,e,f

    [a, b, c, d, e, f] = params

    # calculate new points
    x_new = a * x + b * y + c
    y_new = d * x + e * y + f

    return x_new,y_new

def trans_3R_7params(x,y,z,params=()):
    # parameters in order cx,cy,cz,s,rx,ry,rz (https://en.wikipedia.org/wiki/Helmert_transformation)
    [cx,cy,cz,s,rx,ry,rz] = params

    sec_to_rad = 4.8481e-6
    rx = rx*sec_to_rad
    ry = ry*sec_to_rad
    rz = rz*sec_to_rad


    x_new = cx + (1+s*10e-6)*(x - rz*y + ry*z)

    y_new = cy + (1+s*10e-6)*(rz*x + y - rx*z)
    z_new = cz + (1+s*10e-6)*(-ry*x + rx*y + z)

    # import numpy as np
    # mx = np.array([x,y,z])
    # mc = np.array([cx,cy,cz])
    # mr = np.array([[1,-rz,ry],[rz,1,-rx],[-ry,rx,1]])
    # a = mc + (1+s*0.000001)*np.dot(mr,mx)
    # x_new, y_new, z_new = a

    return x_new,y_new,z_new

def trans_2R_4params(x,y,params):
    C, D, A, B = params

    if D -180 > 0: #Find angles almost 360 deg, because in file it should be negative angle instead of almost 360!
        D = D - 360

    D = D * 2 * math.pi / 360

    x_new = A + C * x - D * y
    y_new = B + D * x + C * y

    return x_new, y_new

def csv_transformation(df_in,from_crs,method,params=None):
    start = datetime.now()

    # Create a Transformation object with all the neccessary data in it.
    ts = SloTransformation(method=method, from_crs=from_crs, params=params)

    x_f,y_f,z_f = utils.recognize_xyz_fields(df_in)

    pts = df_in[[x_f,y_f]].values.tolist()

    pts = [(i.x,i.y) for i in ts.transform(pts)]
    df_in[x_f]  = [i[0] for i in pts]
    df_in[y_f] = [i[1] for i in pts]

    num_pts = len(pts)
    total_seconds = (datetime.now() - start).total_seconds()
    tock_na_minuto = float(num_pts) / total_seconds
    print("Successully transfromed {} points in {:.2f} seconds with a speed of {:.2f} points/s.".format(num_pts,
                                                                                                         total_seconds,
                                                                                                         tock_na_minuto))
    return df_in

def shp_transformation(df_in,from_crs,method,params=None):
    start = datetime.now()

    """
    Function for transformation from d96 to d48 and vice versa with the triangular transformation
    :param df_in:
    :param from_crs:
    :return:
    """

    #Create a Transformation object with all the neccessary data in it.
    ts = SloTransformation(method=method,from_crs=from_crs,params=params)

    num_pts = 0

    if df_in.ix[0, "geometry"].type == "Point":
        pts = df_in["geometry"].values.tolist()
        num_pts += len(pts)

        # poracunaj nove tocke
        pts_out = ts.transform(points=pts)
        df_in["geometry"] = pts_out

    else:
        # change the geometry by applying transformation formula t0 each vertex:
        for i in df_in.index:
            geom = df_in.loc[i, "geometry"]

            if geom.type == "LineString":
                pts = geom.coords[:]
                num_pts += len(pts)

                # poracunaj nove tocke
                pts_out = ts.transform(points=pts)
                df_in.loc[i, "geometry"] = LineString(pts_out)

            elif geom.type == "MultiLineString":
                lines = []
                for line in geom.geoms[:]:
                    pts = line.coords[:]
                    num_pts += len(pts)

                    # poracunaj nove tocke
                    pts_out = ts.transform(points=pts)
                    lines.append(LineString(pts_out))

                df_in.loc[i, "geometry"] = MultiLineString(lines)

            elif geom.type == "Polygon":

                pts = geom.exterior.coords[:]
                num_pts += len(pts)

                # poracunaj nove tocke
                pts_out = ts.transform(points=pts)
                df_in.loc[i, "geometry"] = Polygon(pts_out)
            else:
                raise NotImplementedError(
                    "Found a geometry method {}! Function is implemented only for Point, LineString, "
                    "MultiLineStrings and Polygons.".format(
                        geom.type))

    total_seconds = (datetime.now() - start).total_seconds()
    tock_na_minuto = float(num_pts) / total_seconds
    print("Successully transfromed {} points in {:.2f} seconds with a speed of {:.2f} points/s.".format(num_pts,
                                                                                                         total_seconds,
                                                                                                         tock_na_minuto))

    return df_in


from __future__ import print_function    # (at top of module)
import click
from . import utils,trans
import geopandas as gpd
import os
import pandas as pd
import warnings



@click.command()
@click.option("--to_crs", required=True, type=click.Choice(["d48", "d96"]), help="Coordinate system to transform your data into")
@click.option("--method", default="triangle",  type=click.Choice(["triangle", "24regions"]), help="Transformation method to be used")
@click.option("--params", required=False,default=None, help="Optional argument: semicolon separated manual parameters, required for each transformation method (24regions:4params, triangle:6params,...")
@click.argument("file_in", required=True)
@click.argument("file_out", required=False)
def cli(to_crs,method,params,file_in,file_out):

    if to_crs == 'd48':
        from_crs = "d96"
        to_epsg = 3912
        from_epsg = 379
    else:
        from_crs = "d48"
        to_epsg = 3794
        from_epsg = 3912

    params = [float(i) for i in params.split(";")] if params != None else None


    name, ext = os.path.splitext(os.path.abspath(file_in))

    #Save file!
    if not file_out:
        file_out = name + "_" + to_crs + ext

    # Command line utility tool for two way transformation between old slovenian CRS (D48GK) and new slovenian CRS (D96TM).
    click.echo("Transformating a file {} into {} and saving it to {}...".format(file_in,to_crs,file_out))


    if ext == ".shp":

        #Check file input epsg code (can only be 3912 or 3794 or none!)
        file_epsg = utils.cheesy_slovenian_epsg_checker(file_in)

        if file_epsg == from_epsg:
            pass
        elif file_epsg == None:
            raise IOError("Shapelile can only be in crs d48 (EPSG:3912) or d96 (EPSG:3794)! It didn't find word 'slovene' in prj file!")
        elif file_epsg == to_epsg:
            raise IOError("The input shapefile you've specified ({}) already seems to be in a projection "
                          "that you want to transform it into ({}, epsg: {})! That doesn't make any sense!".format(file_in,to_crs,file_epsg))


        df_in = gpd.read_file(file_in)

        print("\nInput data sample:")
        print(df_in)

        df_out = trans.shp_transformation(df_in=df_in, from_crs=from_crs,method=method,params=params)

        print("\nOutput data sample:")
        print(df_out)

        utils.save_to_shapefile_with_prj(geo_df=df_out, file_out=file_out, epsg=to_epsg)


    elif ext.lower() in [".txt",".csv"]:
        #NO ESPG CHECK...
        warnings.warn("Program can't perform a EPSG check of the input data for csv files for you. "
                      "Let's hope that you specified everything correctly!")

        sep = utils.recognize_csv_separator(file_in)
        header = utils.check_for_csv_header(file_in)
        df_in = pd.read_csv(file_in,sep=sep,header=header)

        print("\nInput data sample:")
        print(df_in)

        df_out = trans.csv_transformation(df_in=df_in, from_crs=from_crs,method=method,params=params)

        print("\nOutput data sample:")
        print(df_out)

        df_out.to_csv(file_out,sep=sep,header=header,index=None)



    else:
        raise IOError("Only shapefiles and csv files are supported input type!")



from __future__ import print_function 

######################################################
#### 1. TRANSFORMING PYTHON LISTS OF POINTS  #########
######################################################

from pysitra import SloTransformation

# List of point that you want converted into d96 via several methods
D48_POINTS = [(500000,100000),(0,0),(650000,200000)]


# Initialize a Triangle Transformation object
ts_triangle = SloTransformation(from_crs="d48",method="triangle")

# Initialize a 24regions transformation object
ts_24region = SloTransformation(from_crs="d48",method="24regions")

# Initialize a affine transformation object with your own parameters
ts_triangle_manual = SloTransformation(from_crs="d48",method="triangle",params="1.00001;0.000040647;-374.668;-0.00002241;1.000006;494.8428".split(";"))

# Note, that seemingly redundant recreation of different transformations as a separate object comes very handy, when you want to
# transform many files/lists at once, so you don't have to perform the expensive transformation object initialization
# for every file/list separately.


# Once you have transformation object initialized, you can use it's .transform() method to transform old points into
# new points quite cheaply:
print("\nTriangle transformation (affine 6parametric):")
print(ts_triangle.transform(D48_POINTS))
print("\n24regions transformation (4parametric):")
print( ts_24region.transform(D48_POINTS))
print("\nTriangle transformation with custom parameters:")
print(ts_triangle_manual.transform(D48_POINTS))

############################################
#### 2. TRANSFORMING FILES WTIH PYTHON##
############################################

from pysitra import shp_transformation,csv_transformation
from pysitra.utils import recognize_csv_separator,check_for_csv_header
import geopandas as gpd
import pandas as pd


# SHAPEFILES:

#read shapefile into GeoDataFrame and transform it and save it as into new shapefile
df_in = gpd.read_file("shapefile_in_d48.shp")
df_out = shp_transformation(df_in,from_crs="d48",method="24regions")
df_out.to_file("shapefile_in_d96.shp")


# ASCII CSVS:
csv_file = "terrain_measurements_in_d48.csv"

sep = recognize_csv_separator(csv_file) #guess the separator type
header = check_for_csv_header(csv_file) #check if file has header

#read csv file into DataFrame, transform them by triangle method with custom parameters and save it to csv.
csv_in = pd.read_csv(csv_file, sep=sep, header=header)
csv_out = csv_transformation(df_in=csv_in, from_crs="d48", method="triangle", params="1.00001;0.000040647;-374.668;-0.00002241;1.000006;494.8428".split(";"))
csv_out.to_file("terrain_measurements_in_d96.csv")





#######################################################################
##### 3. USING LOW LEVEL FUNCTION TO TRANSFORM POINT-BY-POINT  ########
#######################################################################

from pysitra import trans_2R_4params,trans_2R_6params

D48_POINTS = [(500000,100000),(0,0),(650000,200000)]

for point in D48_POINTS:
    # 4parametric transformation with params: scale,rotation,trans_x,trans_y
    x, y = trans_2R_4params(point[0], point[1], params=[0.9999873226,0.0009846750,378.755,-493.382])
    print(x, y)
    # 6parametric transformation with params a,b,c,d,e,f
    x, y = trans_2R_6params(point[0], point[1], params=[1.00001,0.000040647,-374.668,-0.00002241,1.000006,494.8428])
    print(x, y)



from __future__ import print_function
import os
import csv
import warnings
import pandas as pd
import geopandas as gpd
import numpy as np
from shapely.geometry import Point


HOMEDIR = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))


def cheesy_slovenian_epsg_checker(shp):

    #Very silly checker for epsg

    prj_file = os.path.splitext(os.path.abspath(shp))[0] + ".prj"

    with open(prj_file, "r") as f:
        prj_txt = f.read().lower()


    d48_keywords  = ["bessel", "1841", "MGI"]
    d96_keywords  = ["1996","GRS_1980"]

    if not "slove" in prj_txt:
        return None

    for keyword in d48_keywords:
        if keyword in prj_txt:
            return 3912

    for keyword in d96_keywords:
        if keyword in prj_txt:
            return 3794

    else:
        return None

def save_to_shapefile_with_prj(geo_df, file_out, epsg, encoding="utf-8"):
    prj_file = os.path.splitext(os.path.abspath(file_out))[0] + ".prj"

    prj_dict = {
        3912: 'PROJCS["MGI / Slovene National Grid",GEOGCS["MGI",DATUM["D_MGI",SPHEROID["Bessel_1841",6377397.155,299.1528128]],PRIMEM["Greenwich",0],UNIT["Degree",0.017453292519943295]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",15],PARAMETER["scale_factor",0.9999],PARAMETER["false_easting",500000],PARAMETER["false_northing",-5000000],UNIT["Meter",1]]',
        3794: 'PROJCS["Slovenia 1996 / Slovene National Grid",GEOGCS["Slovenia 1996",DATUM["D_Slovenia_Geodetic_Datum_1996",SPHEROID["GRS_1980",6378137,298.257222101]],PRIMEM["Greenwich",0],UNIT["Degree",0.017453292519943295]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",15],PARAMETER["scale_factor",0.9999],PARAMETER["false_easting",500000],PARAMETER["false_northing",-5000000],UNIT["Meter",1]]',
        4326: 'GEOGCS["GCS_WGS_1984",DATUM["D_WGS_1984",SPHEROID["WGS_1984",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["Degree",0.017453292519943295]]',
        3857: 'PROJCS["WGS 84 / Pseudo-Mercator",GEOGCS["Popular Visualisation CRS",DATUM["D_Popular_Visualisation_Datum",SPHEROID["Popular_Visualisation_Sphere",6378137,0]],PRIMEM["Greenwich",0],UNIT["Degree",0.017453292519943295]],PROJECTION["Mercator"],PARAMETER["central_meridian",0],PARAMETER["scale_factor",1],PARAMETER["false_easting",0],PARAMETER["false_northing",0],UNIT["Meter",1]]',
        31258: 'PROJCS["MGI_Austria_GK_M31",GEOGCS["GCS_MGI",DATUM["D_MGI",SPHEROID["Bessel_1841",6377397.155,299.1528128]],PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Transverse_Mercator"],PARAMETER["False_Easting",450000.0],PARAMETER["False_Northing",-5000000.0],PARAMETER["Central_Meridian",13.33333333333333],PARAMETER["Scale_Factor",1.0],PARAMETER["Latitude_Of_Origin",0.0],UNIT["Meter",1.0]]'}

    if epsg not in prj_dict.keys():
        raise Exception("EPSG {} is not in the epsg:wkt_prj dictionary. Add it!".format(epsg))

    with open(prj_file, "w") as f:
        f.write(prj_dict[epsg])

    # save file
    geo_df.to_file(filename=file_out, crs_wkt=prj_dict[epsg], encoding=encoding)

def closest_element_to_given_points(points, shapefile):
    # ce je shapefile argument kot string, potem gre za fajl, ki ga preberi
    if isinstance(shapefile, str):
        shapefile = gpd.read_file(shapefile)

    results = []

    for point in points:

        # je gre za prazno tocko, vrni None
        if not isinstance(point, Point) and (not point[0] or not point[1]):
            results.append(np.NaN)
            continue

        p = Point(point) if not isinstance(point, Point) else point

        distance = 10000000000000
        nearest_object = None

        for i in shapefile.index.tolist():

            ref_object = shapefile.ix[i, "geometry"]

            # ce gre za prazen objekt, vrni None
            if ref_object.is_empty:
                results.append(np.NaN)
                continue

            trenutna_razdalja = p.distance(ref_object)

            if trenutna_razdalja < distance:
                distance = trenutna_razdalja
                nearest_object = i

        results.append(nearest_object)

    return results

def recognize_csv_separator(csv_file):

    # https://stackoverflow.com/questions/3952132/how-do-you-dynamically-identify-unknown-delimiters-in-a-data-file

    with open(csv_file) as f:
            line = f.readline()

    return csv.Sniffer().sniff(line).delimiter

def check_for_csv_header(csv_file):

    with open(csv_file) as f:
            line = f.readline()

    #return direct value for read_csv input argument header
    if any(c.isalpha() for c in set(line)):
        return 0
    else:
        return None

def convert_to_numeric(df):
    for i in df.columns:
        df[i] = pd.to_numeric(df[i], errors="ignore")

    return df

def recognize_xyz_fields(df):

    xyz_candidates = ["e","x","easting","east","vzhod","n","y","northing","north","sever","h","z"]
    x_range = [350000,650000]
    y_range = [10000,300000]
    z_range = [0,10000]

    x,y,z =  None,None,None

    df = convert_to_numeric(df)
    df = df._get_numeric_data() #https://stackoverflow.com/questions/19900202/how-to-determine-whether-a-column-variable-is-numeric-or-not-in-pandas-numpy

    # Check if there are no reasonable column names
    if len(set([str(i).lower() for i in df.columns]).intersection(set(xyz_candidates))) == 0:
        warnings.warn("The dataset specified doesn't seem to have any meaningful column names specifed (e.g: x,y,e,n,easting,...)\n"
                      "Program will try to autodetect x,y and z columns by accepting the first column with appropriate "
                      "value range, starting from the last column...")

        for col in reversed(df.columns):
                mean = df[col].mean()
                if x_range[0] < mean < x_range[-1]:
                    x = col
                elif y_range[0] < mean < y_range[-1]:
                    y = col
                elif z_range[0] < mean < z_range[-1]:
                    z = col
    else:
        for col in reversed(df.columns):
            if col.lower() in xyz_candidates:
                mean = df[col].mean()
                if x_range[0] < mean < x_range[-1]:
                    x = col
                if y_range[0] < mean < y_range[-1]:
                    y = col
                if z_range[0] < mean < y_range[-1]:
                    z = col

    if x == None:
        raise IOError("Couldn't figure out which field represents easting coordinate (tried matching names and guesing by size range.)")

    if y == None:
        raise IOError("Couldn't figure out which field represents northing coordinate (tried matching names and guesing by size range.)")

    if z == None:
        warnings.warn("The specified dataset doesn't seem to have a height attribute specified (h or z field)!")

    return x,y,z




from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='pysitra',
      version='0.3.1',
      description='Python implementation of some popular slovenian transformation methods (SiTra!)',
      long_description=readme(),
      url='https://github.com/marjanmo/pysitra',
      author='Marjan Moderc',
      author_email='marjan.moderc@gmail.com',
      license='MIT',
      packages=['pysitra'],
      install_requires=['pandas','geopandas',"numpy","scipy","shapely","click"],
      include_package_data=True,
      entry_points = {'console_scripts': ['sitra=pysitra.sitra_cli:cli']},
      zip_safe=False)